package at.ac.tuwien.cs.ztimeview.ui;

import java.util.Date;
import java.awt.Color;
import java.awt.geom.Line2D;
import java.awt.Graphics2D;
import java.awt.BasicStroke;
import javax.swing.tree.MutableTreeNode;
import java.beans.PropertyChangeEvent;

import edu.umd.cs.piccolo.util.PPaintContext;
import edu.umd.cs.piccolo.PNode;

import at.ac.tuwien.cs.ztimeview.time.TimeInterval;
import at.ac.tuwien.cs.ztimeview.time.Interval;
import at.ac.tuwien.cs.ztimeview.time.ActivePlanningInterval;

/**
Piccolo class for a PlanningLine 
(LifeLine with hierarchical decompostion and temporal uncertainties)

includes cached loading on demand of children
  
@author Wolfgang Aigner
@version 0.1
  
 created: 2004-06-06
 last change: 2004-06-06
*/
public class PPlanningLine extends PLifeLinePlus {

	/* color of caps */
	final static Color LINE_COLOR = Color.black;
	/* ratio of stroke width to line height */
	final static double STROKE_WIDTH_RATIO = 0.08;
	
	PCap startCap = null; // UI element for start interval
	PCap endCap = null; // UI element for end interval
	PTimeLayer maxDu = null; // UI element for maximum duration bar
	PTimeLayer minDu = null; // UI element for minimum duration bar

	ActivePlanningInterval pinterval = null; // temporal attributes
	
	/** Constructor
	@param aTitle title of the PlanningLine
	@param aInterval temporal attributes
	@param userObject user object for this view */
	public PPlanningLine(String title, ActivePlanningInterval interval, MutableTreeNode userObject) {
		super(title, new TimeInterval(
			new Date(Math.min(interval.getEarliestBegin().getTime(), interval.getStart().getTime())), 
			new Date(Math.max(interval.getLatestEnd().getTime(), interval.getEnd().getTime()))), userObject);
	
		this.pinterval = interval;
		
		this.initialize();
	}

	/** Initialize UI and child elements */
	private void initialize() {
		this.setCollisionFreeLayout(false); // don't use lined up layout
		
		// --- set maximum duration bar ---
		this.maxDu = new PTimeLayer(new TimeInterval(
			this.pinterval.getMaxDuBegin(), this.pinterval.getMaxDuEnd()));
		this.maxDu.setPickable(false);
		this.addChild(0, this.maxDu);
		
		// --- set minimum duration bar ---
		this.minDu = new PTimeLayer(new TimeInterval(
			this.pinterval.getMinDuBegin(), this.pinterval.getMinDuEnd()));
		this.minDu.setDefaultHeight(5);
		this.minDu.setPickable(false);
		this.addChild(1, this.minDu);
		
		// --- set start interval cap ---
		this.startCap = new PCap(new TimeInterval(
			this.pinterval.getEarliestStartingTime(), this.pinterval.getLatestStartingTime()), true);
		this.startCap.setPickable(false);
		this.addChild(startCap);

		// --- set end interval cap ---
		this.endCap = new PCap(new TimeInterval(
			this.pinterval.getEarliestFinishingTime(), this.pinterval.getLatestFinishingTime()), false);
		this.endCap.setPickable(false);
		this.addChild(endCap);
		
		// --- set bounds of children ---
		this.setChildBounds();
	}
	
	/** Property change listener
	@param pce property change event */
	public void propertyChange(PropertyChangeEvent pce) {
		super.propertyChange(pce);
		
		boolean isRelevantBoundsEvent = 
			(PNode.PROPERTY_BOUNDS == pce.getPropertyName() 
			|| PNode.PROPERTY_FULL_BOUNDS == pce.getPropertyName());	
				
		if (isRelevantBoundsEvent) { // if bounds changed
			this.setChildBounds();
		} 		
	}
	
	/** Set bounds of child elements according to line height */
	public void setChildBounds() {
		// --- start cap ---
		if(this.startCap != null) {
			this.startCap.setBounds(this.startCap.getX(), this.startCap.getY(), 
				this.startCap.getWidth(), this.getHeight());
		}
		// --- end cap ---
		if(this.endCap != null) {
			this.endCap.setBounds(this.endCap.getX(), this.endCap.getY(), 
				this.endCap.getWidth(), this.getHeight());
		}
		/* set height of duration bars smaller to be inside caps */
		double inset = this.getHeight() * STROKE_WIDTH_RATIO;
		// --- maximum duration bar ---
		if(this.maxDu != null) {
			this.maxDu.setBounds(this.maxDu.getX(), 0,
				this.maxDu.getWidth(), this.getHeight() - (1.9*inset));
		}
		// --- minimum duration bar ---
		if(this.minDu != null) {
			this.minDu.setBounds(this.minDu.getX(), 0,
				this.minDu.getWidth(), this.getHeight() - (1.9*inset));
		}
	}
	
	/** Expand the PlanningLine and show its children */
	public void expand() {	
		super.expand();

		if(userObject.getChildCount() > 0) {
			// --- set visibility ---
			this.startCap.setVisible(false);
			this.endCap.setVisible(false);
			this.maxDu.setVisible(false);
			this.minDu.setVisible(false);
		}
	}
	
	/** Collapse PlanningLine (hide children) */
	public void collapse() {
		super.collapse();
		
		// --- set visibility ---
		this.startCap.setVisible(true);
		this.endCap.setVisible(true);
		this.maxDu.setVisible(true);
		this.minDu.setVisible(true);
	}
	
	/** Layout children of current layer */
	public void layoutChildren() {
		super.layoutChildren();
		
		// --- set vertical offset of duration bars to be centered vertically ---
		double inset = this.getHeight() * STROKE_WIDTH_RATIO;
		
		if(!this.isExpanded()) {
			if(this.maxDu != null) {
				this.maxDu.setOffset(this.maxDu.getOffset().getX(), 
					this.minDu.getOffset().getY() + inset);
			}
			
			if(this.minDu != null) {
				this.minDu.setOffset(this.minDu.getOffset().getX(), 
					this.minDu.getOffset().getY() + inset);
			}
		}
		
	}
	
	/** Paint the PlanningLine
	@param aPaintContext paint context for drawing */
	public void paint(PPaintContext aPaintContext) {
		// --- set colors ---
		this.maxDu.setPaint(((Color) this.getPaint()).brighter());
		this.minDu.setPaint(this.getPaint());
		// --- semantic zoom for label ---
		if(!isExpanded()) {
			double s = aPaintContext.getScale();
			this.label.setVisible((this.label.getHeight()*s) > MIN_VISIBLE_FONT_HEIGHT);
		}

	}

	
	//////////////////// INNER CLASSES ////////////////////

	/** Inner class that represents a start / end interval cap
	@author Wolfgang Aigner
	@version 0.1
	  
	 created: 2004-06-06
	 last change: 2004-06-06
	 */
	class PCap extends PTimeLayer {
		
		boolean left = true; // is left cap?
		
		/** Constructor */
		public PCap() {
			this(new TimeInterval(new Date(), new Date()), true);
		}
		
		/** Constructor
		@param aInterval temporal extent of cap
		@param left is left cap? */
		public PCap(Interval aInterval, boolean left) {
			super(aInterval);
			this.left = left;
		}
		
		/** Paint the PlanningLine
		@param aPaintContext paint context for drawing */
		public void paint(PPaintContext aPaintContext) {
			Graphics2D g2 = aPaintContext.getGraphics();
			double strokeWidth = this.getHeight() * STROKE_WIDTH_RATIO;
			double inset = strokeWidth/2;
			// --- set line color and width ---
			g2.setPaint(LINE_COLOR);
			g2.setStroke(new BasicStroke((float) strokeWidth));
			// --- draw upper line ---
			double x1 = (this.getWidth() > 2*inset ? this.getWidth() - inset : inset);
			Line2D.Double upperLine = new Line2D.Double(
				inset, inset, x1, inset);
			g2.draw(upperLine);	
			// --- draw lower line ---
			Line2D.Double lowerLine = new Line2D.Double(
				inset, this.getHeight() - inset,
				x1, this.getHeight() - inset);
			g2.draw(lowerLine);
			// --- draw vertical line ---
			double x2 = (left ? inset : this.getWidth() - inset);
			if(this.getWidth() < inset) {
				x2 = inset;
			}
			Line2D.Double vertLine = 
				new Line2D.Double(x2, inset, x2, this.getHeight() - inset);
			g2.draw(vertLine);
		}
	}

}
