package at.ac.tuwien.cs.ztimeview.ui;

import java.util.Date;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.TimeZone;
import java.util.Vector;
import java.lang.String;
import java.lang.Long;
import java.lang.Math;
import java.text.SimpleDateFormat;

/**
defines axis and calculates interval
for ticks according to current start and
end value
  
@author Martin Tomitsch
@version 0.3
  
 created: 2004-06-05
 last change: 2004-06-07
*/
public class PTimeScale {

	/* --- Display granularity constants --- */
	public final static int AUTO = 0;
	public final static int MILLISECONDS = 1;
	public final static int SECONDS = 2;
	public final static int MINUTES = 3;
	public final static int HOURS = 4;
	public final static int DAYS = 5;
	public final static int WEEKS = 6;
	public final static int MONTHS = 7;
	public final static int YEARS = 8;
	
	/* --- Date format strings --- */
	public final static String MILLISECOND_FORMAT = "ss.SSS";
	public final static String MILLISECOND_FORMAT_EXTENDED = "ss";
	
	public final static String SECOND_FORMAT = "ss";
	public final static String SECOND_FORMAT_EXTENDED = "HH:mm";
	
	public final static String MINUTE_FORMAT = "mm";
	public final static String MINUTE_FORMAT_EXTENDED = "HH:mm";
	
	public final static String HOUR_FORMAT = "HH:mm";
	public final static String HOUR_FORMAT_EXTENDED = "d.M.";
	
	public final static String DAY_FORMAT = "d.M.";
	public final static String DAY_FORMAT_EXTENDED = "d.M.yy";

	public final static String WEEK_FORMAT = "d.M.";
	public final static String WEEK_FORMAT_EXTENDED = "d.M.yy";
	
	public final static String MONTH_FORMAT = "MMM";
	public final static String MONTH_FORMAT_EXTENDED = "MMM.yy";
	
	public final static String YEAR_FORMAT = "yyyy";
	
	/* --- distance of small ticks within main ticks --- */
	public final static int SECONDS_SMALL_TICKS = 250;
	public final static int MINUTES_SMALL_TICKS = 15;
	public final static int HOURS_SMALL_TICKS = 60;
	public final static int DAYS_SMALL_TICKS = 6;
	public final static int WEEKS_SMALL_TICKS = 1;
	public final static int MONTHS_SMALL_TICKS = 7;
	public final static int YEARS_SMALL_TICKS = 3;
	
	/** display granularity */
	int displayGranularity = 0;
	
	/** displayed time interval */
	private Date start=null, end=null;
		
	private int axisLength, axisX; //length and x position of axis in pixel
	private int LABELSIZE = 50; //length of label text in pixel
	private int TICKSOFFSET = 50; //minimum offset between two ticks so that next granularity can be added
	
	/** constructor */
	public PTimeScale(long start, long end, int axisLength, int axisX) {
		this.start = new Date(start); 
		this.end = new Date(end); 
		this.axisLength = axisLength;
		this.axisX = axisX;
		this.setAutoDisplayGranularity(); //set granularity for scale
	}
	
	/////////////////////// METHODS //////////////////////
	
	/** calculates the tick intervals and returns them as a list */
	Vector calculateTicks() {
		Vector ticks = new Vector();
		PTick tick;
		
		long interval = end.getTime() - start.getTime(); //interval between dates in ms
		double currentInterval=0;
		int calendarGranularity = 0;
		
		switch(displayGranularity) {
			case MILLISECONDS:
				currentInterval = interval;
				calendarGranularity = Calendar.MILLISECOND;
				break;
			case SECONDS:
				currentInterval = (double)interval/1000;
				calendarGranularity = Calendar.SECOND;
				break;
			case MINUTES:
				currentInterval = (double)interval/(1000*60);
				calendarGranularity = Calendar.MINUTE;
				break;
			case HOURS:
				currentInterval = (double)interval/(1000*60*60);
				calendarGranularity = Calendar.HOUR;
				break;
			case DAYS:
				currentInterval = (double)interval/86400000; //number of days being displayed on the timescale
				calendarGranularity = Calendar.DATE; //needed for Calendar add method
				break;
			case WEEKS:
				currentInterval = (double)interval/(86400000*7);
				calendarGranularity = Calendar.WEEK_OF_YEAR;
				break;
			case MONTHS:
				currentInterval = (double)interval/86400000.0d/30.5d; 
				calendarGranularity = Calendar.MONTH;
				break;
			case YEARS:
				currentInterval = (double)interval/86400000d/365d;
				calendarGranularity = Calendar.YEAR;
				break;
		}
		
		//System.err.print(calendarGranularity + ": " + currentInterval + " (" + interval + ")\n"); //debug
		
		double scaleInterval = (double)axisLength / (double)currentInterval;
		double previousLabelPos=LABELSIZE/(-2); //position of previous label
		
		GregorianCalendar actTime = new GregorianCalendar();
		actTime.setTime(start);
		
		//find first aligned value for the current granularity
		boolean aligned = false;
		switch(displayGranularity) {
			case MILLISECONDS:
				break;
			case SECONDS:
				for(int i = 0; i < 1000 && !aligned; i += SECONDS_SMALL_TICKS) {
					aligned = (actTime.get(Calendar.MILLISECOND) == i);
				}
				if(!aligned) {
					actTime.add(Calendar.MILLISECOND,
								(1000 - actTime.get(Calendar.MILLISECOND)) % SECONDS_SMALL_TICKS);
				}
				break;
			case MINUTES:
				for(int i = 0; i < 60 && !aligned; i += MINUTES_SMALL_TICKS) {
					aligned = (actTime.get(Calendar.SECOND) == i);
				}
				if(!aligned) {
					actTime.add(Calendar.SECOND,
								(60 - actTime.get(Calendar.SECOND)) % MINUTES_SMALL_TICKS);
				}
				actTime.set(Calendar.MILLISECOND, 0);
				break;
			case HOURS:
				for(int i = 0; i < 60 && !aligned; i += HOURS_SMALL_TICKS) {
					aligned = (actTime.get(Calendar.MINUTE) == i);
				}
				if(!aligned) {
					actTime.add(Calendar.MINUTE,
								(60 - actTime.get(Calendar.MINUTE)) % HOURS_SMALL_TICKS);
				}
				actTime.set(Calendar.SECOND, 0);
				actTime.set(Calendar.MILLISECOND, 0);
				break;
			case DAYS:
				for(int i = 0; i < 12 && !aligned; i += DAYS_SMALL_TICKS) {
					aligned = (actTime.get(Calendar.HOUR) == i);
				}
				if(!aligned) {
					actTime.add(Calendar.HOUR, 
								(12 - actTime.get(Calendar.HOUR)) % DAYS_SMALL_TICKS);
				}
				actTime.set(Calendar.MINUTE, 0);
				actTime.set(Calendar.SECOND, 0);
				actTime.set(Calendar.MILLISECOND, 0);
				
				break;
			case WEEKS:
				aligned = actTime.get(Calendar.HOUR_OF_DAY) == 0;
				if(!aligned) {
					actTime.set(Calendar.DATE, actTime.get(Calendar.DATE) + 1);
				}
				actTime.set(Calendar.AM_PM, 0);
				actTime.set(Calendar.MINUTE, 0);
				actTime.set(Calendar.SECOND, 0);
				actTime.set(Calendar.MILLISECOND, 0);
				
				break;
			case MONTHS:
				aligned = actTime.get(Calendar.HOUR_OF_DAY) == 0;
				if(!aligned) {
					actTime.set(Calendar.DATE, actTime.get(Calendar.DATE) + 1);
				}
				actTime.set(Calendar.AM_PM, 0);
				actTime.set(Calendar.MINUTE, 0);
				actTime.set(Calendar.SECOND, 0);
				actTime.set(Calendar.MILLISECOND, 0);
				
				break;
			case YEARS:
				for(int i = 0; i < 24 && !aligned; i += YEARS_SMALL_TICKS) {
					aligned = (actTime.get(Calendar.MONTH) == i);
				}
				if(!aligned) {
					actTime.add(Calendar.MONTH, 
								(24 - actTime.get(Calendar.MONTH)) % MONTHS_SMALL_TICKS);
				}
				actTime.set(Calendar.DATE, 1);
				actTime.set(Calendar.HOUR_OF_DAY, 0);
				actTime.set(Calendar.MINUTE, 0);
				actTime.set(Calendar.SECOND, 0);
				actTime.set(Calendar.MILLISECOND, 0);
				
				break;
		}
		
		//change x position (starting position) of timescale if date was aligned
		axisX += (actTime.getTime().getTime() - start.getTime()) * ((double)axisLength/(double)interval);
			
		int currentTickLength;
		int lastLabelIndex=0;
		
		for(int i = 0; i < currentInterval; i++) {
			tick = new PTick();
			currentTickLength = (int)(i*scaleInterval); //length of current (one) interval in pixel
			tick.setCanvasX(axisX+currentTickLength);
			if (((tick.getCanvasX()-previousLabelPos > LABELSIZE*1.5)) && ((axisLength-tick.getCanvasX()) > LABELSIZE/2)) { 
				actTime.add(calendarGranularity,i-lastLabelIndex); //set current time
				lastLabelIndex = i;
				SimpleDateFormat formatter = this.getDateFormatter(actTime); //format current time
				tick.setLabel(formatter.format(actTime.getTime()));
				//set flag for different formatter on timescale
				if ((displayGranularity == MILLISECONDS && formatter.toPattern().compareTo(MILLISECOND_FORMAT_EXTENDED) == 0) ||
					(displayGranularity == SECONDS && formatter.toPattern().compareTo(SECOND_FORMAT_EXTENDED) == 0) ||
					(displayGranularity == MINUTES && formatter.toPattern().compareTo(MINUTE_FORMAT_EXTENDED) == 0) ||
					(displayGranularity == HOURS && formatter.toPattern().compareTo(HOUR_FORMAT_EXTENDED) == 0) ||
					(displayGranularity == DAYS && formatter.toPattern().compareTo(DAY_FORMAT_EXTENDED) == 0) ||
					(displayGranularity == WEEKS && formatter.toPattern().compareTo(WEEK_FORMAT_EXTENDED) == 0) ||
					(displayGranularity == MONTHS && formatter.toPattern().compareTo(MONTH_FORMAT_EXTENDED) == 0)) {
					tick.setMainLabel(true);
				}
				previousLabelPos = tick.getCanvasX();
			}
			else
				tick.setLabel("");
			
			ticks.add(tick);	
			
			//if interval between two ticks of the current granularity is large enough, 
			//add three ticks for next granularity
			if (scaleInterval > TICKSOFFSET) {
				int n;
				if (displayGranularity == WEEKS) n = 7;
				else n = 4;
				for (int j=1; j < n; j++) {
					tick = new PTick();
					tick.setCanvasX(axisX+currentTickLength+j*(scaleInterval/n));
					tick.setSmallerGranularity(true);
					ticks.add(tick);
				}
			}
		}
		
		return ticks;
	}
	
	/** converts an element of type Date to String
		@param date
		@return String representing the given date */
	public String dateToLabel(long date) {
		GregorianCalendar c = new GregorianCalendar();
		c.setTime(new Date(date));
		return Integer.toString(c.get(c.DAY_OF_MONTH))
			+ '.' 
			+ Integer.toString(c.get(c.MONTH) + 1) // return value is pointer into field, starting with 0
			+ '.' 
			+ Integer.toString(c.get(c.YEAR));
	}
	
	/** Set the display granularity according to the given interval */
	private void setAutoDisplayGranularity() {
		if(start != null && end != null) {
			long interval = end.getTime() - start.getTime();
			GregorianCalendar intervalCal = new GregorianCalendar();
			intervalCal.setTimeZone(TimeZone.getTimeZone("GMT"));
			intervalCal.setTime(new Date(interval));
			
			if(intervalCal.get(Calendar.YEAR) - 1970 > 1) this.displayGranularity = YEARS;
			else if(intervalCal.get(Calendar.YEAR) - 1970 == 1 
				|| intervalCal.get(Calendar.MONTH) > 4) this.displayGranularity = MONTHS;
			else if(intervalCal.get(Calendar.MONTH) >= 1
				|| intervalCal.get(Calendar.DATE) > 10) this.displayGranularity = WEEKS;
			else if(intervalCal.get(Calendar.DATE) > 3) this.displayGranularity = DAYS;
			else if(intervalCal.get(Calendar.DATE) == 3
				|| intervalCal.get(Calendar.HOUR_OF_DAY) > 1) this.displayGranularity = HOURS;
			else if(intervalCal.get(Calendar.HOUR_OF_DAY) == 1
				|| intervalCal.get(Calendar.MINUTE) >= 1) this.displayGranularity = MINUTES;
			else if(intervalCal.get(Calendar.SECOND) >= 1) this.displayGranularity = SECONDS;
			else this.displayGranularity = MILLISECONDS;
		}
	}
	
	/** Get date formatter for specified granularity and instant
		@param actTime instant to get formatter for */
	private SimpleDateFormat getDateFormatter(Calendar actTime) {
		SimpleDateFormat formatter = null;
		
		switch(displayGranularity) {
			case MILLISECONDS:
				if(actTime.get(Calendar.MILLISECOND) == 0) {
					formatter = new SimpleDateFormat (MILLISECOND_FORMAT_EXTENDED);
				}
				else {
					formatter = new SimpleDateFormat (MILLISECOND_FORMAT);
				}	
				break;
			case SECONDS:
				if(actTime.get(Calendar.SECOND) == 0) {
					formatter = new SimpleDateFormat (SECOND_FORMAT_EXTENDED);
				}
				else {
					formatter = new SimpleDateFormat (SECOND_FORMAT);
				}	
				break;
			case MINUTES:
				if(actTime.get(Calendar.MINUTE) == 0) {
					formatter = new SimpleDateFormat (MINUTE_FORMAT_EXTENDED);
				}
				else {
					formatter = new SimpleDateFormat (MINUTE_FORMAT);
				}	
				break;
			case HOURS:
				if(actTime.get(Calendar.HOUR_OF_DAY) == 0) {
					formatter = new SimpleDateFormat (HOUR_FORMAT_EXTENDED);
				}
				else {
					formatter = new SimpleDateFormat (HOUR_FORMAT);
				}	
				break;				
			case DAYS:
				if(actTime.get(Calendar.DATE) == 1) {
					formatter = new SimpleDateFormat (DAY_FORMAT_EXTENDED);
				}
				else {
					formatter = new SimpleDateFormat (DAY_FORMAT);
				}
				break;
			case WEEKS:
				if(actTime.get(Calendar.MONTH) == 0 && actTime.get(Calendar.DATE) == 1) {
					formatter = new SimpleDateFormat (WEEK_FORMAT_EXTENDED);
				}
				else {
					formatter = new SimpleDateFormat (WEEK_FORMAT);
				}
				break;
			case MONTHS:
				if(actTime.get(Calendar.MONTH) == 0 && actTime.get(Calendar.DATE) == 1) {
					formatter = new SimpleDateFormat (MONTH_FORMAT_EXTENDED);
				}
				else {
					formatter = new SimpleDateFormat (MONTH_FORMAT);
				}
				break;
			case YEARS:
				formatter = new SimpleDateFormat (YEAR_FORMAT);
				
				break;
		}
		
		return formatter;
	}
	
	//////////////////// GET/SET METHODS ////////////////////
	/** sets the range for the axis
	@param start lower date 
	@param end upper date */
	void setRange(long start, long end) {
		this.start = new Date(start); 
		this.end = new Date(end); 
	}
	
	/** sets length of axis
	@param axisLength */
	void setLength(int axisLength) {
		this.axisLength = axisLength;
	}
	
}