package at.ac.tuwien.cs.ztimeview.test;

import java.io.File;
import java.util.Date;

import at.ac.tuwien.cs.ztimeview.core.Task;
import at.ac.tuwien.cs.ztimeview.core.PlanningTask;
import at.ac.tuwien.cs.ztimeview.time.TimeInterval;

/**
Class for generating test data (task hierarchy)

@author Wolfgang Aigner
@version 0.4
 
 created: 2004-05-19
 last change: 2004-06-07
*/
public class TaskGenerator {
	
	final static int LEN_MULTIPLIER = 100000; // multiplication factor for length
	
	// ignore these files
	final static String IGNORE_FILE = ".DS_STORE";
	
	/** Generate a task hierarchy from files in a given directory
	Tasks are generated in the following way:
		file/dir -> task(name: filename, start: lastModified, end: lastModified+length*factor)
		dir -> hierarchy; files in dir are child tasks
	@param dir directory to start from
	@param planning create planningTasks randomly
	*/
	public static Task generateTaskFromDir(File dir, boolean planning) {

		Date start = new Date(dir.lastModified());
		Date end = new Date(dir.lastModified() + dir.length()*LEN_MULTIPLIER);
		
		/* 50% of created tasks should have temporal uncertainties */
		boolean createPlanningTask = (planning ? Math.random() > 0.5 : false);
		Task t = null;
		
		if(createPlanningTask) {
			t = new PlanningTask();
		} else {
			t = new Task();
		}
		
		t.setName(dir.getName());
		
		// --- add children ---
		if(dir.isDirectory()) {
			try {
				File[] files = dir.listFiles();
				boolean first = true;
		
				for(int i=0; i<files.length; i++) {	
					if(!files[i].getName().equalsIgnoreCase(IGNORE_FILE)) {
						Task child = generateTaskFromDir(files[i], planning);
						t.add(child);
						
						// set parent interval according to children
						if(first) {
							start = child.getStart();
							end = child.getEnd();
							first = false;
						} else {
							if(child.getStart().getTime() < start.getTime()) { 
								// child starts earlier -> modify parent start
								start = child.getStart();
							}
							if(child.getEnd().getTime() > end.getTime()) {
								// child ends later -> modify parent end
								end = child.getEnd();		
							}
						}
					}
				}
			} catch(java.lang.SecurityException e) {
				System.err.println(e.getMessage());		
			}			
		}
		
		t.setStart(start);
		t.setEnd(end);

		if(createPlanningTask) { // add temporal uncertainties
			addPlanningAttributes((PlanningTask) t);
		}
		
		return t;

	}
	
	/** Add random temporal attributes to a planning task 
	@param pt task to add attributes to */
	private static void addPlanningAttributes(PlanningTask pt) {
		// --- set earliest starting time ---
		pt.setEarliestStartingTime(pt.getStart());
		// --- set latest finishing time ---
		pt.setLatestFinishingTime(pt.getEnd());
		
		TimeInterval ti = new TimeInterval(pt);
		long intervalMax = ti.getDuration() / 3;
		
		//--- set latest starting time ---
		Date lst = new Date(pt.getStart().getTime() + (long) (Math.random() * intervalMax));
		pt.setLatestStartingTime(lst);
		// --- set earliest finishing time ---
		Date eft = new Date(pt.getEnd().getTime() - (long) (Math.random() * intervalMax));
		pt.setEarliestFinishingTime(eft);

		// --- set maximum duration ---
		double factor1 = Math.random();
		double factor2 = Math.random();
		
		Date est = pt.getEarliestStartingTime();
		Date lft = pt.getLatestFinishingTime();
		
		long maxDuOld = pt.getMaximumDuration().longValue();
		long maxDiff1 = maxDuOld - (eft.getTime() - est.getTime());
		long maxDiff2 = maxDuOld - (lft.getTime() - lst.getTime());
		long diffMax = Math.min(maxDiff1, maxDiff2);
		
		long maxDu = maxDuOld - ((long) (Math.random() * diffMax));
		
		pt.setMaximumDuration(new Long(maxDu));
		
		// --- set minimum duration ---
		long minDuOld = pt.getMinimumDuration().longValue();
		long diffMin = maxDu - minDuOld;
		
		long minDu = minDuOld + ((long) (Math.random() * diffMin));
		minDu = Math.min(minDu, eft.getTime() - est.getTime());
		minDu = Math.min(minDu, lft.getTime() - lst.getTime());

		pt.setMinimumDuration(new Long(minDu));

	}
	
	/** Test method
	Generates task hierarchy and prints it to screen
	*/
	public static void main(String args[]) {	
		Task t = new Task();

		try {
			t = TaskGenerator.generateTaskFromDir(new File(args[1]), 
				args[0].equalsIgnoreCase("planningline"));
		} catch(java.lang.NullPointerException e) {
			System.err.println(e.getMessage());
		}
		
		t.printHierarchy();
	}
}