package at.ac.tuwien.cs.ztimeview.time;

import java.util.*;
import at.ac.tuwien.cs.ztimeview.time.PlanningInterval;
import java.text.MessageFormat;

/**
Class representing a time annotation as used in Asbru 7.3.
  
@author Wolfgang Aigner
@version 0.8
  
 created: 2003-02-17
 last change: 2003-03-27
*/
public class TimeAnnotation implements PlanningInterval {
	// --- error messages ---
	final static String ERR_LSS_BEFORE_ESS = "The latest starting shift can not be "
		+ "before earliest starting shift. Enter a latest starting shift greater or equal to "
		+ "{0,date,medium}, {0,time,HH:mm:ss.SSS}";
	final static String ERR_LFS_BEFORE_EFS = "The latest finishing shift can not be "
		+ "before earliest finishing shift. Enter a latest finishing shift greater or equal to "
		+ "{0,date,medium}, {0,time,HH:mm:ss.SSS}";
	final static String ERR_EFS_BEFORE_ESS = "The earliest finishing shift can not be "
		+ "before earliest starting shift. Enter a earliest finishing shift greater or equal to "
		+ "{0,date,medium}, {0,time,HH:mm:ss.SSS}";
	final static String ERR_LFS_BEFORE_LSS = "The latest finishing shift can not be "
		+ "before latest starting shift. Enter a latest finishing shift greater or equal to "
		+ "{0,date,medium}, {0,time,HH:mm:ss.SSS}";
	final static String ERR_MAXDU_SHORTER_MINDU = "The maximum duration can not be "
		+ "shorter than the minimum duration. Enter a maximum duration longer than"
		+ "{0,number} ms";
	final static String ERR_MINDU_SHORTER_EFSLSS = "The minimum duration can not be "
		+ "shorter than EFS - LSS. Enter a minimum duration longer than "
		+ "{0,number} ms";
	final static String ERR_MINDU_LONGER_EFSESS = "The minimum duration can not be "
		+ "longer than EFS - ESS. Enter a minimum duration shorter or equal than "
		+ "{0,number} ms";
	final static String ERR_MINDU_LONGER_LFSLSS = "The minimum duration can not be "
		+ "longer than LFS - LSS. Enter a minimum duration shorter or equal than "
		+ "{0,number} ms";
	final static String ERR_MAXDU_SHORTER_EFSESS = "The maximum duration can not be "
		+ "shorter than EFS - ESS. Enter a maximum duration longer than "
		+ "{0,number} ms";
	final static String ERR_MAXDU_SHORTER_LFSLSS = "The maximum duration can not be "
		+ "shorter than LFS - LSS. Enter a maximum duration longer than "
		+ "{0,number} ms";
	final static String ERR_MAXDU_LONGER_LFSESS = "The maximum duration can not be "
		+ "longer than LFS - ESS. Enter a maximum duration shorter than "
		+ "{0,number} ms";
		
	
	public final static String MILLISECONDS = "ms";
	
	// --- constant for not assigned values ---
	public final static Long NA = null;

	// --- attributes in milliseconds ---
	Long ess = NA; // earliest starting shift
	Long lss = NA; // latest starting shift
	Long efs = NA; // earliest finishing shift
	Long lfs = NA; // latest finishing shift
	Long minDu = NA; // minimum duration
	Long maxDu = NA; // maximum duration
	TimeAnnotationReference ref = null; // reference point
		
	//////////////////// CONSTRUCTORS ////////////////////
	
	/** Default constructor */
	public TimeAnnotation() throws IllegalArgumentException {
		this(null, NA, NA, NA, NA, NA, NA);
	}
	
	/** Constructor 1
	@param ess earliest starting shift in milliseconds
	@param lss latest starting shift in milliseconds
	@param efs earliest finishing shift in milliseconds
	@param lfs latest finishing shift in milliseconds
	@param minDu minimum duration in milliseconds
	@param maxDu maximum duration in milliseconds */
	public TimeAnnotation(Long ess, Long lss, Long efs, Long lfs, 
		Long minDu, Long maxDu) throws IllegalArgumentException {
		this(new ReferencePoint(), ess, lss, efs, lfs, minDu, maxDu);
	}
	
	/** Constructor 2 
	@param ref reference point
	@param ess earliest starting shift in milliseconds
	@param lss latest starting shift in milliseconds
	@param efs earliest finishing shift in milliseconds
	@param lfs latest finishing shift in milliseconds
	@param minDu minimum duration in milliseconds
	@param maxDu maximum duration in milliseconds */
	public TimeAnnotation(TimeAnnotationReference ref, Long ess, Long lss, 
		Long efs, Long lfs, Long minDu, Long maxDu) throws IllegalArgumentException {
		super();
		
		this.ref = ref;
		this.ess = ess;
		this.lss = lss;
		this.efs = efs;
		this.lfs = lfs;
		this.minDu = minDu;
		this.maxDu = maxDu;
		
		// check integrity of time annotation attributes
		TimeAnnotation.checkAttributes(ref, ess, lss, efs, lfs, minDu, maxDu); 	
	}
	
	//////////////////// STATIC METHODS ////////////////////
	
	/** Check integrity of time annotation attributes */
	private static void checkAttributes(TimeAnnotationReference ref, Long ess, 
		Long lss, Long efs, Long lfs, Long minDu, Long maxDu) 
		throws IllegalArgumentException {
		
		// --- check time annotation attributes ---
		if(lss != NA && ess != NA
			&& lss.compareTo(ess) < 0) { // LSS < ESS
			throw new IllegalArgumentException(MessageFormat.format(
				ERR_LSS_BEFORE_ESS,  new Object[]{new Date(ess.longValue())}));
		}
		else if( lfs != NA && efs != NA
			&& lfs.compareTo(efs) < 0) {  // LFS < EFS
			throw new IllegalArgumentException(MessageFormat.format(
				ERR_LFS_BEFORE_EFS, new Object[]{new Date(efs.longValue())} ));
		}
		else if(efs != NA && ess != NA
			&& efs.compareTo(ess) < 0) {  // EFS < ESS
			throw new IllegalArgumentException(MessageFormat.format(
				ERR_EFS_BEFORE_ESS, new Object[]{new Date(ess.longValue())} ));
		}
		else if(lfs != NA && lss != NA
			&& lfs.compareTo(lss) < 0) {  // LFS < LSS
			throw new IllegalArgumentException(MessageFormat.format(
				ERR_LFS_BEFORE_LSS, new Object[]{new Date(lss.longValue())} ));
		}
		else if(minDu != NA && efs != NA && lss != NA
			&& minDu.longValue() < (efs.longValue() - lss.longValue())) {  // MinDu < EFS-LSS
			throw new IllegalArgumentException(MessageFormat.format(
				ERR_MINDU_SHORTER_EFSLSS, new Object[]{new Long(efs.longValue() - lss.longValue())} ));
		}
		else if(minDu != NA && efs != NA && ess != NA
			&& minDu.longValue() > (efs.longValue() - ess.longValue())) {  // MinDu > EFS-ESS
			throw new IllegalArgumentException(MessageFormat.format(
				ERR_MINDU_LONGER_EFSESS, new Object[]{new Long(efs.longValue() - ess.longValue())} ));
		}
		else if(minDu != NA && lfs != NA && lss != NA
			&& minDu.longValue() > (lfs.longValue() - lss.longValue())) {  // MinDu > LFS-LSS
			throw new IllegalArgumentException(MessageFormat.format(
			ERR_MINDU_LONGER_LFSLSS, new Object[]{new Long(lfs.longValue() - lss.longValue())} ));
		}
		else if(maxDu != NA && efs != NA && ess != NA
			&& maxDu.longValue() < (efs.longValue() - ess.longValue())) {  // MaxDu < EFS-ESS
			throw new IllegalArgumentException(MessageFormat.format(
				ERR_MAXDU_SHORTER_EFSESS, new Object[]{new Long(efs.longValue() - ess.longValue())} ));
		}
		else if(maxDu != NA && lfs != NA && lss != NA
			&& maxDu.longValue() < (lfs.longValue() - lss.longValue())) {  // MaxDu < LFS-LSS
			throw new IllegalArgumentException(MessageFormat.format(
				ERR_MAXDU_SHORTER_LFSLSS, new Object[]{new Long(lfs.longValue() - lss.longValue())} ));
		}
		else if(maxDu != NA && lfs != NA && ess != NA
			&& maxDu.longValue() > (lfs.longValue() - ess.longValue())) {  // MaxDu > LFS-ESS
			throw new IllegalArgumentException(MessageFormat.format(
				ERR_MAXDU_LONGER_LFSESS, new Object[]{new Long(lfs.longValue() - ess.longValue())} ));
		}
	}
	
	//////////////////// METHODS OF ITimeAnnotation INTERFACE ////////////////////
	
	/** Set earliest starting time
	@param est earliest starting time */
	public void setEarliestStartingTime(Date est) {
		this.setESS(new Long(est.getTime() - ref.getAbsoluteValue().getTime()));
	}
	
	/** Get earliest starting time
	@return earliest starting time */
	public Date getEarliestStartingTime() {
		Date est = null;
		if(ref != null && ess != NA) {
			est = new Date(ref.getAbsoluteValue().getTime() + ess.longValue());
		}
		return est;
	}
	
	/** Set latest starting time
	@param lst latest starting time */
	public void setLatestStartingTime(Date lst) {
		this.setLSS(new Long(lst.getTime() - ref.getAbsoluteValue().getTime()));
	}
	
	/** Get latest starting time
	@return latest starting time */
	public Date getLatestStartingTime() {
		Date lst = null;
		if(ref != null && lss != NA) {
			lst = new Date(ref.getAbsoluteValue().getTime() + lss.longValue());
		}
		return lst; 
	}
	
	/** Set earliest finishing time
	@param eft earliest finishing time */
	public void setEarliestFinishingTime(Date eft) {
		this.setEFS(new Long(eft.getTime() - ref.getAbsoluteValue().getTime()));
	}
	
	/** Get earliest finishing time
	@return earliest finishing time */
	public Date getEarliestFinishingTime() {
		Date eft = null;
		if(ref != null && efs != NA) {
			eft = new Date(ref.getAbsoluteValue().getTime() + efs.longValue());
		}
		return eft; 
	}
	
	/** Set latest finishing time
	@param lft latest finishing time */
	public void setLatestFinishingTime(Date lft) {
		this.setLFS(new Long(lft.getTime() - ref.getAbsoluteValue().getTime()));
	}
	
	/** Get latest finishing time
	@return latest finishing time */
	public Date getLatestFinishingTime() {
		Date lft = null;
		if(ref != null && lfs != NA) {
			lft = new Date(ref.getAbsoluteValue().getTime() + lfs.longValue());
		}
		return lft; 
	}
	
	/** Get minimum duration
	@return minimum duration in milliseconds */
	public Long getMinimumDuration() {
		Long minimumDuration = null;
		
		if(minDu != NA) {
			minimumDuration = minDu;
		}
		else { // try to calculate minimum duration
			// --- get lower bound ---
			Long lowerBound = null;
			if(lss != NA) {
				lowerBound = lss;
			}
			else if(ess != NA) {
				lowerBound = ess;
			}
			
			// --- get upper bound ---
			Long upperBound = null;
			if(efs != NA) {
				upperBound = efs;
			}
			else if(lfs != NA) {
				upperBound = lfs;
			}
			
			// --- calculate minimum duration ---
			if(lowerBound != NA && upperBound != NA) {
				minimumDuration = new Long(upperBound.longValue() - lowerBound.longValue());
			}
		}
		
		return minimumDuration;
	}
	
	/** Get minimum duration
	@return minimum duration in milliseconds */
	public Long getMaximumDuration()	 {
		Long maximumDuration = null;
		
		if(maxDu != NA) {
			maximumDuration = maxDu;
		}
		else { // try to calculate maximum duration
			// --- get lower bound ---
			Long lowerBound = null;
			if(ess != NA) {
				lowerBound = ess;
			}
			else if(lss != NA) {
				lowerBound = lss;
			}
			
			// --- get upper bound ---
			Long upperBound = null;
			if(lfs != NA) {
				upperBound = lfs;
			}
			else if(efs != NA) {
				upperBound = efs;
			}
			
			// --- calculate maximum duration ---
			if(lowerBound != NA && upperBound != NA) {
				maximumDuration = new Long(upperBound.longValue() - lowerBound.longValue());
			}
		}
		
		return maximumDuration;
	}
	
	/** Get earliest instant of time annotation */
	public Date getEarliestBegin() {
		Date begin = null;
		
		if(this.getEarliestStartingTime() != null) {
			begin = this.getEarliestStartingTime();
		}
		else if(this.getMaxDuBegin() != null) {
			begin = getMaxDuBegin();
		}
		else if(this.getMinDuBegin() != null) {
			begin = getMinDuBegin();
		}
		else if(this.getLatestStartingTime() != null) {
			begin = this.getLatestStartingTime();
		}
		else if(this.getEarliestFinishingTime() != null) {
			begin = this.getEarliestFinishingTime();
		}
		else if(this.getLatestFinishingTime() != null) {
			begin = this.getLatestFinishingTime();
		}
		
		return begin;
	}
	
	/** Get latest instant of time annotation */
	public Date getLatestEnd() {
		Date end = null;
		
		if(this.getLatestFinishingTime() != null) {
			end = this.getLatestFinishingTime();
		}
		else if(this.getMaxDuBegin() != null) {
			end = new Date();
			end.setTime(this.getMaxDuBegin().getTime() 
				+ this.getMaximumDuration().longValue());
		}
		else if(this.getMinDuBegin() != null) {
			end = new Date();
			end.setTime(this.getMinDuBegin().getTime() 
				+ this.getMinimumDuration().longValue());
		}
		else if(this.getEarliestFinishingTime() != null) {
			end = this.getEarliestFinishingTime();
		}
		else if(this.getLatestStartingTime() != null) {
			end = this.getLatestStartingTime();
		}
		else if(this.getEarliestStartingTime() != null) {
			end = this.getEarliestStartingTime();
		}
				
		return end;
	}
	
	/** Get beginning of maxDu interval */
	public Date getMaxDuBegin() {
		Date t1 = null;
		Date t2 = null;
		Date begin = null;
		
		if(this.getEarliestStartingTime() != null) {
			t1 = this.getEarliestStartingTime();
		}
		else if(this.getLatestStartingTime() != null) {
			t1 = this.getLatestStartingTime();
		}
		
		if(this.getLatestFinishingTime() != null) {
			t2 = this.getLatestFinishingTime();
		}
		else if(this.getEarliestFinishingTime() != null) {
			t2 = this.getEarliestFinishingTime();
		}
		
		if(t1 != null && t2 != null && this.getMaximumDuration() != null) {
			long durDiff = this.getMaximumDuration().longValue()
				- (t2.getTime() - t1.getTime());
			begin = new Date();
			begin.setTime(t1.getTime() - (long) (durDiff / 2));
		}

		return begin;
	}
	
	/** Get end of maxDu interval */
	public Date getMaxDuEnd() {
		Date maxDuBegin = this.getMaxDuBegin();
		Date maxDuEnd = new Date();
		
		maxDuEnd.setTime(maxDuBegin.getTime() 
			+ this.getMaximumDuration().longValue());
			
		return maxDuEnd;
	}
	
	/** Get beginning of minDu interval */
	public Date getMinDuBegin() {
		Date t1 = null;
		Date t2 = null;
		Date begin = null;
		
		if(this.getLatestStartingTime() != null) {
			t1 = this.getLatestStartingTime();
		}
		else if(this.getEarliestStartingTime() != null) {
			t1 = this.getEarliestStartingTime();
		}
		
		if(this.getEarliestFinishingTime() != null) {
			t2 = this.getEarliestFinishingTime();
		}
		else if(this.getLatestFinishingTime() != null) {
			t2 = this.getLatestFinishingTime();
		}
		
		if(t1 != null && t2 != null && this.getMinimumDuration() != null) {
			long durDiff = this.getMinimumDuration().longValue()
				- (t2.getTime() - t1.getTime());
			begin = new Date();
			begin.setTime(t1.getTime() - (long) (durDiff / 2));
		}
		
		return begin;
	}
	
	/** Get end of minDu interval */
	public Date getMinDuEnd() {
		Date minDuBegin = this.getMinDuBegin();
		Date minDuEnd = new Date();
		
		minDuEnd.setTime(minDuBegin.getTime() 
			+ this.getMinimumDuration().longValue());
			
		return minDuEnd;
	}

	
	//////////////////// GET/SET METHODS ////////////////////
	
	/** Set reference point
	@param ref reference point */
	public void setRef(TimeAnnotationReference ref) throws IllegalArgumentException {
		TimeAnnotation.checkAttributes(ref, this.ess, this.lss, this.efs, 
			this.lfs, this.minDu, this.maxDu);
		this.ref = ref;
	}
	
	/** Get reference point
	@return reference point */
	public TimeAnnotationReference getRef() {
		return this.ref;
	}
	
	/** Set earliest starting shift
	@param ess earliest starting shift in milliseconds */
	public void setESS(Long ess) throws IllegalArgumentException {
		TimeAnnotation.checkAttributes(this.ref, ess, this.lss, this.efs, 
			this.lfs, this.minDu, this.maxDu);
		this.ess = ess;
	}
	
	/** Get earliest starting shift
	@return earliest starting shift in milliseconds */
	public Long getESS() {
		return this.ess;
	}
	
	/** Set latest starting shift
	@param ess latest starting shift in milliseconds */
	public void setLSS(Long lss) throws IllegalArgumentException {
		TimeAnnotation.checkAttributes(this.ref, this.ess, lss, this.efs, 
			this.lfs, this.minDu, this.maxDu);
		this.lss = lss;
	}
	
	/** Get latest starting shift
	@return latest starting shift in milliseconds */
	public Long getLSS() {
		return this.lss;
	}
	
	/** Set earliest finishing shift
	@param ess earliest finishing shift in milliseconds */
	public void setEFS(Long efs) throws IllegalArgumentException {
		TimeAnnotation.checkAttributes(this.ref, this.ess, this.lss, efs, 
			this.lfs, this.minDu, this.maxDu);
		this.efs = efs;
	}
	
	/** Get earliest finishing shift
	@return earliest finishing shift in milliseconds */
	public Long getEFS() {
		return this.efs;
	}
	
	/** Set latest finishing shift
	@param ess latest finishing shift in milliseconds */
	public void setLFS(Long lfs) throws IllegalArgumentException {
		TimeAnnotation.checkAttributes(this.ref, this.ess, this.lss, this.efs, 
			lfs, this.minDu, this.maxDu);
		this.lfs = lfs;
	}
	
	/** Get latest finishing shift
	@return latest finishing shift in milliseconds */
	public Long getLFS() {
		return this.lfs;
	}
	
	/** Set minimum duration
	@param minDu minimum duration in milliseconds */
	public void setMinDu(Long minDu) throws IllegalArgumentException {
		TimeAnnotation.checkAttributes(this.ref, this.ess, this.lss, this.efs, 
			this.lfs, minDu, this.maxDu);
		this.minDu = minDu;
	}
	
	/** Set minimum duration
	@param minDu minimum duration in milliseconds */
	public void setMinimumDuration(Long minDu) throws IllegalArgumentException {
		setMinDu(minDu);
	}
	
	/** Get latest finishing shift
	@return minimum duration in milliseconds */
	public Long getMinDu() {
		return this.minDu;
	}
	
	/** Set maximum duration
	@param maxDu maximum duration in milliseconds */
	public void setMaximumDuration(Long maxDu) throws IllegalArgumentException {
		setMaxDu(maxDu);
	}
	
	/** Set maximum duration
	@param maxDu maximum duration in milliseconds */
	public void setMaxDu(Long maxDu) throws IllegalArgumentException {
		TimeAnnotation.checkAttributes(this.ref, this.ess, this.lss, this.efs, 
			this.lfs, this.minDu, maxDu);
		this.maxDu = maxDu;
	}
	
	/** Get latest finishing shift
	@return maximum duration in milliseconds */
	public Long getMaxDu() {
		return this.maxDu;
	}

}
