package at.ac.tuwien.cs.ztimeview.time;

import java.util.*;
import java.text.MessageFormat;

/**
Class representing a time interval and implementing the Allen's relations
  
@author Wolfgang Aigner
@version 0.6
  
 created: 2003-02-25
 last change: 2003-04-01
*/
public class TimeInterval implements Interval {
	// --- error messages ---
	final static String END_BEFORE_START = "The end can not be "
		+ "before the start of the interval. Enter an end time greater or equal to "
		+ "{0,date,medium}, {0,time,HH:mm:ss.SSS}";
	
	final static String START_AFTER_END = "The start can not be "
		+ "after the end of the interval. Enter an start time less or equal to "
		+ "{0,date,medium}, {0,time,HH:mm:ss.SSS}";
		
	public final static String MILLISECONDS = "ms";

	// --- start and end of interval ---
	Date start = null;
	Date end = null;
		
	//////////////////// CONSTRUCTORS ////////////////////
	
	/** Default constructor */
	public TimeInterval() throws IllegalArgumentException {
		this(null, null);
	}
	
	/** Constructor
	@param interval time interval */
	public TimeInterval(Interval interval) throws IllegalArgumentException {
		this(interval.getStart(), interval.getEnd());
	}
	
	/** Constructor
	@param start start time of interval
	@param end end time of interval */
	public TimeInterval(Date start, Date end) throws IllegalArgumentException {
		super();
		this.setStart(start);
		this.setEnd(end);
	}
	
	////////////////////  STATIC METHODS ////////////////////
	
	/** Determine if the provided intervals are equal
	@param i interval to compare */
	public static boolean areEqual(Interval i1, Interval i2) {
		return (i1.getStart().equals(i2.getStart()) && i1.getEnd().equals(i2.getEnd()));
	}
	
	////////////////////  ALLEN'S RELATIONS ////////////////////
	
	/** Determine if this interval equals the given one
	@param i interval to compare */
	public boolean equals(TimeInterval i) {
		return (this.start.equals(i.getStart()) && this.end.equals(i.getEnd()));
	}
	
	/** Determine if this interval is before the given one
	@param i interval to compare */
	public boolean before(TimeInterval i) {
		return (this.end.before(i.getStart()));
	}
	
	/** Determine if this interval is after the given one
	@param i interval to compare */
	public boolean after(TimeInterval i) {
		return i.before(this);
	}
	
	/** Determine if this interval meets the given one
	@param i interval to compare */
	public boolean meets(TimeInterval i) {
		return (this.end.equals(i.getStart()));
	}
	
	/** Determine if this interval is met by the given one
	@param i interval to compare */
	public boolean metBy(TimeInterval i) {
		return i.meets(this);
	}

	/** Determine if this interval overlaps the given one
	@param i interval to compare */
	public boolean overlaps(TimeInterval i) {
		return (this.start.before(i.getStart()) 
			&& this.end.after(i.getStart()) 
			&& this.end.before(i.getEnd()));
	}
	
	/** Determine if this interval is overlapped by the given one
	@param i interval to compare */
	public boolean overlappedBy(TimeInterval i) {
		return i.overlaps(this);
	}
	
	/** Determine if this interval starts the given one
	@param i interval to compare */
	public boolean starts(TimeInterval i) {
		return (this.start.equals(i.getStart()) 
			&& this.end.after(i.getStart()) 
			&& this.end.before(i.getEnd()));
	}
	
	/** Determine if this interval is started by the given one
	@param i interval to compare */
	public boolean startedBy(TimeInterval i) {
		return i.starts(this);
	}
	
	/** Determine if this interval is during the given one
	@param i interval to compare */
	public boolean during(TimeInterval i) {
		return (this.start.after(i.getStart()) 
			&& this.end.before(i.getEnd()));
	}
	
	/** Determine if this interval contains the given one
	@param i interval to compare */
	public boolean contains(TimeInterval i) {
		return i.during(this);
	}
	
	/** Determine if this interval finishes the given one
	@param i interval to compare */
	public boolean finishes(TimeInterval i) {
		return (this.end.equals(i.getEnd()) 
			&& this.start.after(i.getStart()) 
			&& this.start.before(i.getEnd()));
	}
	
	/** Determine if this interval is finished by the given one
	@param i interval to compare */
	public boolean finishedBy(TimeInterval i) {
		return i.finishes(this);
	}
	
	//////////////////// METHODS ////////////////////
	
	/** Determine if given instant is within interval
	@param t instant to test */
	public boolean contains(Date t) {
		return (!t.before(this.start) && !t.after(this.end));
	}
	
	/** Determine if given instant is before interval
	@param t instant to test */
	public boolean after(Date t) {
		return (t.before(this.start));
	}
	
	/** Determine if given instant is after interval
	@param t instant to test */
	public boolean before(Date t) {
		return (t.after(this.end));
	}

	/** Get the duration of the interval
	@return duration in milliseconds */
	public long getDuration() {
		return (this.end.getTime() - this.start.getTime());
	}
	
	/** Get the duration of the interval
	@param interval Interval to calculate duration for
	@return duration in milliseconds */
	public static long getDuration(Interval interval) {
		return (interval.getEnd().getTime() - interval.getStart().getTime());
	}
	
	//////////////////// GET/SET METHODS ////////////////////
	
	/** Set start time
	@param start start time of plan */
	public void setStart(Date start) throws IllegalArgumentException {
		if(this.end != null && start != null && this.end.before(start)) {
			throw new IllegalArgumentException(MessageFormat.format(
				START_AFTER_END, 
				new Object[]{this.end} ));
		}
		this.start = start;
	}
	
	/** Get start time
	@return start time of plan */
	public Date getStart() {
		return start;
	}
	
	/** Set end of plan
	@param end end of plan */
	public void setEnd(Date end) {
		if(this.start != null && end != null && this.start.after(end)) {
			throw new IllegalArgumentException(MessageFormat.format(
				END_BEFORE_START, 
				new Object[]{this.start} ));
		}
		this.end = end;
	}
	
	/** Get end of plan
	@return end of plan */
	public Date getEnd() {
		return this.end;
	}

}
