package at.ac.tuwien.cs.ztimeview.ui;

import java.util.Date;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.LinkedList;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.awt.Graphics2D;
import java.awt.Paint;

import edu.umd.cs.piccolo.PNode;
import edu.umd.cs.piccolo.PLayer;
import edu.umd.cs.piccolo.util.PBounds;
import edu.umd.cs.piccolo.util.PPaintContext;

import at.ac.tuwien.cs.ztimeview.time.Interval;
import at.ac.tuwien.cs.ztimeview.time.TimeInterval;

/**
Piccolo Layer for Timed Objects

@author Wolfgang Aigner
@version 0.4
  
 created: 2004-05-20
 last change: 2004-06-07
*/
public class PTimeLayer extends PLayer implements Interval {
	
	// temporal extent
	Interval interval = null;
	
	// lines for laying out children
	LinkedList layoutLines = null; 
	
	// precision of display width[px] = duration[ms]*PRECISION
	static double PRECISION = 0.000001; 
	
	// space between rows of children (px)
	static int VERTICAL_SPACING = 5;
	
	// default height of line
	double defaultHeight = 30;
	
	// always use default height
	boolean alwaysUseDefaultHeight = true;
	
	// apply collision free layout theme?
	boolean collisionFreeLayout = true;
	
	/** Default constructor
	create time layer */
	public PTimeLayer() {
		this(new TimeInterval(new Date(), new Date()));
	}
	
	/** Constructor
	@param aInterval temporal extent of time layer */
	public PTimeLayer(Interval aInterval) {
		super();
		this.interval = aInterval;
		this.setExtent();
	}
	
 
	/** Add a child node to the current node
	overrides standard method
	adds setting of bounds when child is added
	adds layout lines for overlap prevention
	@param index array index for child to insert
	@param child child node to insert
	*/
	public void addChild(int index, PNode child) {
		super.addChild(index, child);
			
		if(child instanceof Interval) {
			TimeInterval childInterval = new TimeInterval((Interval) child);
			
			if(this.layoutLines == null) { // no child intervals yet -> create layout aid
				this.layoutLines = new LinkedList();	
			}
			
			boolean inserted = false; // child element has been inserted
			boolean conflict = false; 	// child element overlaps with 
										//other element in current line
			
			ListIterator lineIterator = layoutLines.listIterator();
			LinkedList currLine = null;
			TimeInterval currInterval = null;
			
			while(lineIterator.hasNext() && !inserted) { // investigate layout lines
				currLine = (LinkedList) lineIterator.next();
				conflict = false;
				
				ListIterator elemIterator = currLine.listIterator(); 
				while(elemIterator.hasNext() && !inserted && !conflict) {
				// elements of current line
					
					currInterval = new TimeInterval((Interval) elemIterator.next());
					
					if(childInterval.before(currInterval)) {
						// --- insert here ---
						currLine.add(elemIterator.nextIndex()-1, child);
						inserted = true;
					} else {
						// --- conflict? -> try next line ---
						conflict = !childInterval.after(currInterval);
					}		
					// --- try next element in line ---
				}
				
				if(!inserted && !conflict) { // insert at the end of the current line
					currLine.add(child);
					inserted = true;	
				}
			}
	     	
			if(!inserted) { // add new line and insert element
				LinkedList newLine = new LinkedList();
				newLine.add(child);
				layoutLines.add(newLine);	
			}
 	
		}
		
		this.setExtent();

	}
	
	public void removeAllChildren() {
		super.removeAllChildren();
		this.layoutLines = null;	
	}
	
	/** Set the extent (width) of the layer 
	according to temporal extent and precision */
	public void setExtent() {
		if(interval != null) {
			// set width according to duration
			// set height according to enclosed elements
			TimeInterval ti = new TimeInterval(interval);
			double width = ti.getDuration()*PRECISION;
			
			double height = this.defaultHeight;
			if(!this.alwaysUseDefaultHeight) {
				//double newHeight = this.getUnionOfChildrenBounds(null).getHeight();
				double newHeight = this.getFullBoundsReference().getHeight();
				height = (newHeight > this.defaultHeight ? 
					newHeight : this.defaultHeight);
			}
			this.setBounds(this.getX(), this.getY(), width, height);
		}	
	}
	
	/** Layout children of current layer
	children are laid out according to their temporal position within the layer
	and avoiding overlapping
	*/
	public void layoutChildren() {				
		double xOffset = 0;
		double yOffset = 0;
		double lineHeight = 0;
		
		if(layoutLines != null) {
			PNode currElem = null;
			for(Iterator lineIterator = layoutLines.listIterator(); lineIterator.hasNext(); ) {
				LinkedList currLine = (LinkedList) lineIterator.next();
				for(Iterator elemIterator = currLine.listIterator(); elemIterator.hasNext(); ) {
					currElem = (PNode) elemIterator.next();
					if(currElem instanceof Interval) {
						TimeInterval startDiff = 
							new TimeInterval(this.getStart(), ((Interval) currElem).getStart());
									
						xOffset = (startDiff.getDuration() * this.getWidth()) / 
							(new TimeInterval(this)).getDuration();
							
						currElem.setOffset(xOffset, yOffset - currElem.getY());
						
						if(currElem.getHeight() > lineHeight) {
							lineHeight = currElem.getFullBoundsReference().getHeight();	
						}
					}	
				}
				if(this.collisionFreeLayout) {
					yOffset += lineHeight + VERTICAL_SPACING;
					lineHeight = 0;
				}
			}
		}
	}
	
	//////////////////// GET/SET METHODS ////////////////////
	
	/** Set start time
	@param start start time */
	public void setStart(Date start) throws IllegalArgumentException {
		if(this.interval == null) { 
			this.interval = new TimeInterval();
		}
		interval.setStart(start);
	}
	
	/** Get start time
	@return start time */
	public Date getStart() {
		return interval.getStart();
	}
	
	/** Set end 
	@param end end time */
	public void setEnd(Date end) throws IllegalArgumentException {
		if(this.interval == null) { 
			this.interval = new TimeInterval();
		}
		interval.setStart(end);
	}
	
	/** Get end 
	@return end time */
	public Date getEnd() {
		return interval.getEnd();
	}
	
	/** Set collision free layout
	@param collisionFree collision free layout on/off */
	public void setCollisionFreeLayout(boolean collisionFree) {
		this.collisionFreeLayout = collisionFree;
	}
	
	/** Is collision free layout turned on?
	@return collision free layout turned on */
	public boolean getCollisionFreeLayout() {
		return this.collisionFreeLayout;
	}
	
	/** Set the default height 
	@param height default height */
	public void setDefaultHeight(double height) {
		this.defaultHeight = height;
		this.setExtent();
	}

	/** Get the default height
	@return default height */
	public double getDefaultHeight() {
		return this.defaultHeight;
	}
	
	/** Always use the default height
	@param always use default height */
	public void setAlwaysUseDefaultHeight(boolean alwaysUseDefaultHeight) {
		this.alwaysUseDefaultHeight = alwaysUseDefaultHeight;
		this.setExtent();
	}

}
