/**
*  Class for generating test data (task hierarchy)
*  
*  @author Wolfgang Aigner
*  @version 0.3
*  
*  created: 2004-05-19
*  last change: 2004-06-06
*/

package at.ac.tuwien.cs.ztimeview.test;

import java.io.File;
import java.util.Date;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.DefaultMutableTreeNode;

import at.ac.tuwien.cs.ztimeview.core.Task;
import at.ac.tuwien.cs.ztimeview.core.Milestone;
import at.ac.tuwien.cs.ztimeview.core.TimeElement;
import at.ac.tuwien.cs.ztimeview.time.TimeInterval;

public class TaskGenerator {
	
	final static long LEN_MULTIPLIER = 100000; // multiplication factor for length
	
	final static long MIN_LEN_FOR_TASK = 500; // minimal length to be a task
	
	/** Generate a task hierarchy from files in a given directory
	Tasks are generated in the following way:
		file/dir -> task(name: filename, start: lastModified, end: lastModified+length*factor)
		dir -> hierarchy; files in dir are child tasks
	@param dir directory to start from
	*/
	public static TimeElement generateTimeElementFor(File dir) {

		TimeElement te = null;
		Date start = new Date(dir.lastModified());
		Date end = new Date(dir.lastModified() + dir.length()*LEN_MULTIPLIER);
		TimeInterval interval = new TimeInterval(start, end);
		
		if(interval.getDuration() > (MIN_LEN_FOR_TASK * LEN_MULTIPLIER) 
			|| dir.isDirectory()) { // generate task
		
			te = new Task();
			((Task) te).setName(dir.getName());
			
			if(dir.isDirectory()) {
				try {
					File[] files = dir.listFiles();
			
					for(int i=0; i<files.length; i++) {	
						TimeElement child = generateTimeElementFor(files[i]);
						((DefaultMutableTreeNode) te).add((MutableTreeNode)child);
						
						// set parent interval according to children
						if(i==0) {
							if(child instanceof Task) {
								start = ((Task) child).getStart();
								end = ((Task) child).getEnd();
							} else { // Milestone
								start = ((Milestone) child).getTime();
								end = ((Milestone) child).getTime();
							}
						} else {
							if(child instanceof Task) {
								if(((Task) child).getStart().getTime() < start.getTime()) { 
									// child starts earlier -> modify parent start
									start = ((Task) child).getStart();
								}
								if(((Task) child).getEnd().getTime() > end.getTime()) {
									// child ends later -> modify parent end
									end = ((Task) child).getEnd();		
								}
							} else { // Milestone
								if(((Milestone) child).getTime().getTime() < start.getTime()) { 
									// child starts earlier -> modify parent start
									start = ((Milestone) child).getTime();
								}
								if(((Milestone) child).getTime().getTime() > end.getTime()) {
									// child ends later -> modify parent end
									end = ((Milestone) child).getTime();		
								}
							}
						}
					}
				} catch(java.lang.SecurityException e) {
					System.err.println(e.getMessage());		
				}			
			}
			
			((Task) te).setStart(start);
			((Task) te).setEnd(end);
		} else { // generate Milestone
			te = new Milestone(dir.getName(), start);
		}
		
		return te;
	}
	
	/** Test method
	Generates task hierarchy and prints it to screen
	*/
	public static void main(String args[]) {	
		TimeElement t = null;
		
		try {
			t = TaskGenerator.generateTimeElementFor(new File(args[0]));
		} catch(java.lang.NullPointerException e) {
			System.err.println(e.getMessage());
		}
		
		t.printHierarchy(0);
	}
}