/**
*  Piccolo class for a LifeLine+ 
*  (LifeLine with hierarchical decompostion)
*
*  includes cached loading on demand of children
*  
*  @author Wolfgang Aigner
*  @version 0.3
*  
*  created: 2004-06-02
*  last change: 2004-06-05
*/
package at.ac.tuwien.cs.ztimeview.ui;

import java.util.Date;
import java.util.Enumeration;
import java.util.ArrayList;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Color;
import java.awt.event.InputEvent;
import javax.swing.tree.MutableTreeNode;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.GeneralPath;

import edu.umd.cs.piccolo.PNode;
import edu.umd.cs.piccolo.nodes.PText;
import edu.umd.cs.piccolo.nodes.PPath;
import edu.umd.cs.piccolo.util.PPaintContext;
import edu.umd.cs.piccolo.util.PBounds;
import edu.umd.cs.piccolo.event.PBasicInputEventHandler;
import edu.umd.cs.piccolo.event.PInputEvent;
import edu.umd.cs.piccolo.event.PInputEventFilter;

import at.ac.tuwien.cs.ztimeview.time.Interval;
import at.ac.tuwien.cs.ztimeview.time.TimeInterval;
import at.ac.tuwien.cs.ztimeview.util.PTimeElementFactory;

public class PLifeLinePlus extends PTimeLayer implements PropertyChangeListener, PTimeElement {
	
	// color of summary line
	final static Color SUMMARY_LINE_COLOR = Color.lightGray;
	// color of expand handle
	final static Color EXPAND_HANDLE_COLOR = Color.lightGray;
	// color of expand handle border
	final static Color EXPAND_HANDLE_BORDER_COLOR = Color.white;
	
	// height of line = HEIGHT_TO_WIDTH_RATIO * width of line
	final static double HEIGHT_TO_WIDTH_RATIO = 0.04;
	// height of summary line = SUMMARY_LINE_TO_HEIGHT_RATIO * height of line
	final static double SUMMARY_LINE_TO_HEIGHT_RATIO = 0.25;
	// horizontal space between elements
	final static double HORIZONTAL_SPACING_RATIO = 0.5;
	// vertical space between elements
	final static double VERTICAL_SPACING_RATIO = 0.25;
	// inset for expand handle
	final static double EXPAND_HANDLE_INSET_RATIO = 0.25;
	
	/* minimal height for semantic zoom */
	final static int MIN_VISIBLE_HEIGHT = 5;
	// minimal height of line
	final static int MIN_HEIGHT = 30;
	
	// all LifeLines should have the same height?
	final static boolean ALL_SAME_HEIGHT = false;
	
	// default height of line
	double defaultHeight = 30;
	
	// user object - contains user data
	MutableTreeNode userObject = null;
	
	// parent of this lifeline
	PLifeLinePlus parentLine = null;
	// cache for child elements
	ArrayList childCache = null;
	
	// title of the line
	String title = null;
	
	// indicates if children are currently visible
	boolean childrenVisible = false;
	
	// UI elements
	PText label = null; // label conatining line title
	PTimeLayer childLayer = null; // container for children lines
	PSummaryLine summaryLine = null; // summary line
	PPath expandHandle = null; // expand handle (triangle)
	
	/** Default constructor */
	public PLifeLinePlus() {
		this("LifeLine+", new TimeInterval(new Date(), new Date()), null);
	}
	
	/** Constructor
	@param aTitle title of the LifeLine+
	@param aInterval temporal extent
	@param userObject user object for this view */
	public PLifeLinePlus(String title, Interval interval, MutableTreeNode userObject) {
		super(interval);
		this.userObject = userObject;
		this.title = title;
		
		this.initialize();
	}
		
	/** Initialize UI and child elements */
	private void initialize() {
		// --- set height ---
		double newDefaultHeight = this.getWidth() * HEIGHT_TO_WIDTH_RATIO;
		this.defaultHeight = (newDefaultHeight > MIN_HEIGHT ? newDefaultHeight : MIN_HEIGHT);
		this.setExtent();
		
		// --- add property change listener ---
		this.addPropertyChangeListener(this);
		
		// --- add tooltip ---
		this.addClientProperty("tooltip", this.title);
		
		// --- add label ---
		this.label = new PText(this.title);
		this.label.setPickable(false);
		this.addChild(this.label);	
		
		// --- add expand handle and summary line if there are sub-elements ---
		if(userObject != null && userObject.getChildCount() > 0) {
		
			// --- add summary line ---
			this.summaryLine = new PSummaryLine();
			this.setSummaryLineBounds();
			
			// set color
			this.summaryLine.setPaint(SUMMARY_LINE_COLOR);
			
			// add tooltip
			this.summaryLine.addClientProperty("tooltip", "^ " + this.title);
			
			// add mouse listener for collapsing event
			PBasicInputEventHandler collapseEventHandler = 
				new PBasicInputEventHandler() {
				public void mousePressed(PInputEvent event) {
					PNode n = event.getPickedNode();
					PLifeLinePlus pickedLine = (PLifeLinePlus) n.getParent();
					
					if(pickedLine != null && !event.isHandled()) {
						pickedLine.collapse();
						event.setHandled(true);
					}
				};
			};
			collapseEventHandler.setEventFilter(
				new PInputEventFilter(InputEvent.BUTTON1_MASK));
			this.summaryLine.addInputEventListener(collapseEventHandler);
			
			// add summaryLine to lifeline
			this.addChild(summaryLine);
			this.summaryLine.setVisible(false);
	
			// --- add child container ---
			this.childLayer = new PTimeLayer(this);
			this.childLayer.addClientProperty("tooltip", this.title);
			this.addChild(this.childLayer);
			
			// --- add expand handle ---
			this.expandHandle = new PPath();
			
			// set colors
			this.expandHandle.setPaint(EXPAND_HANDLE_COLOR);
			this.expandHandle.setStrokePaint(EXPAND_HANDLE_BORDER_COLOR);
			
			// set geometry (triangle)
			this.setExpandHandleBounds();
			this.expandHandle.append(
				new Line2D.Double(0, 0, 0, this.expandHandle.getHeight()), true);
			this.expandHandle.append(
				new Line2D.Double(0, this.expandHandle.getHeight(), 
				this.expandHandle.getWidth(), this.expandHandle.getHeight() / 2), true);
			this.expandHandle.closePath();
			
			// add tooltip
			this.expandHandle.addClientProperty("tooltip", "> " + this.title);
			
			// add mouse listener for expanding event
			PBasicInputEventHandler expandEventHandler = 
				new PBasicInputEventHandler() {
				public void mousePressed(PInputEvent event) {
					PNode n = event.getPickedNode();
					PLifeLinePlus pickedLine = (PLifeLinePlus) n.getParent();
					
					if(pickedLine != null && !event.isHandled()) {
						pickedLine.expand();
						event.setHandled(true);
					}
				};
			};
			expandEventHandler.setEventFilter(
				new PInputEventFilter(InputEvent.BUTTON1_MASK));
			this.expandHandle.addInputEventListener(expandEventHandler);
			
			// add expand handle to lifeline
			this.addChild(this.expandHandle);
		}
	}
	
	/** Set bounds of the expand handle according to line bounds */
	private void setExpandHandleBounds() {
		double inset = this.getHeight() * EXPAND_HANDLE_INSET_RATIO;
		double size = this.getHeight() - (2 * inset);
		this.expandHandle.setBounds(0, 0, size , size);
		this.expandHandle.setOffset(inset, inset);
	}
	
	/**  Set bounds of the summary line according to line bounds*/
	private void setSummaryLineBounds() {
		this.summaryLine.setBounds(0, 0, this.getWidth(), 
			this.getHeight() * SUMMARY_LINE_TO_HEIGHT_RATIO);
		this.summaryLine.setOffset(0, 0);
	}
	
	/** Set height of line according to ratio or children */
	public void setExtent() {
		this.setLabelSize(0.0f); // reset font size of label
		
		super.setExtent();
		//double defaultHeight = this.getWidth() * HEIGHT_TO_WIDTH_RATIO;
		double height = (this.getUnionOfChildrenBounds(null).getHeight() > defaultHeight ? 
			this.getUnionOfChildrenBounds(null).getHeight() : defaultHeight);
		this.setBounds(this.getX(), this.getY(), this.getWidth(), height);
	}
	
	/** Property change listener
	@param pce property change event */
	public void propertyChange(PropertyChangeEvent pce) {
		boolean isRelevantBoundsEvent = 
			(PNode.PROPERTY_BOUNDS == pce.getPropertyName() 
			|| PNode.PROPERTY_FULL_BOUNDS == pce.getPropertyName());	
				
		if (isRelevantBoundsEvent) { // if bounds changed
			// set font size of label according to height of LifeLine
			this.setLabelSize((float)(this.getHeight()*0.7));
			
			if(this.expandHandle != null) {
				this.setExpandHandleBounds();
			}
		} 		
	}
	
	/** Layout children of current layer */
	public void layoutChildren() {
		if(this.isExpanded()) {
			this.childLayer.setOffset(0, this.summaryLine.getHeight()
				+ this.summaryLine.getHeight() * VERTICAL_SPACING_RATIO);
		}
		
		if(this.expandHandle != null) {
			this.label.setOffset(
				this.expandHandle.getWidth() + 
				this.expandHandle.getWidth() * HORIZONTAL_SPACING_RATIO * 2, 0);
		}
	}
	
	/** Paint LifeLine+
	Paints background if line is collapsed. If the line is expanded, only the
	children will be paint.
	@param aPaintContext paint context for drawing */
	public void paint(PPaintContext aPaintContext) {
		if(!isExpanded()) {
			double s = aPaintContext.getScale();
			this.label.setVisible((this.getHeight()*s) > MIN_VISIBLE_HEIGHT);
			super.paint(aPaintContext);
		}
	}
	
	/** Expand the LifeLine+ and show its children */
	public void expand() {	
		Object childElem = null;

		if(userObject.getChildCount() > 0) {
			// --- add children ---
			if(this.childCache != null) {
				// take children from cache
				this.childLayer.addChildren(this.childCache);
			} else {
				// create child lines
				for(Enumeration child = userObject.children(); child.hasMoreElements();) {
					childElem = child.nextElement();
					PTimeElement childView = 
						PTimeElementFactory.getPTimeElementFor(childElem);
					childView.setParentLine(this);
					
					if(ALL_SAME_HEIGHT) {
						childView.setDefaultHeight(this.defaultHeight);
					}
					
					if(childView instanceof PEvent) {
						((PEvent) childView).setBounds(0, 0, this.getHeight(), this.getHeight());
					}
					
					this.childLayer.addChild((PNode)childView);
				}
				// fill in cache
				this.childCache = new ArrayList(this.childLayer.getChildrenReference());
			}
			
			// --- set size ---
			this.setExtent();
			if(this.parentLine != null) {
				this.parentLine.setExtent();
			}
			
			// --- set visibility ---
			this.label.setVisible(false);
			this.summaryLine.setVisible(true);
			this.summaryLine.setPickable(true);
			this.expandHandle.setVisible(false);
			this.expandHandle.setPickable(false);
			
			this.childrenVisible = true;
		}
	}
	
	/** Collapse LifeLine+ (hide children) */
	public void collapse() {
		childLayer.removeAllChildren();
		
		// --- set size ---
		childLayer.setExtent();
		this.setExtent();
		if(this.parentLine != null) {
			this.parentLine.setExtent();
		}
		
		// --- set visibility ---
		this.label.setVisible(true);
		this.summaryLine.setVisible(false);
		this.summaryLine.setPickable(false);
		this.expandHandle.setVisible(true);
		this.expandHandle.setPickable(true);
		
		this.childrenVisible = false;
	}
	
	/** Determines wheter the children are visible
	@return children visible */
	public boolean isExpanded() {
		return this.childrenVisible;	
	}

	/** Set the font size of the label
	@param size font size */
	private void setLabelSize(float size) {
		if(this.label != null) {
			this.label.setFont(this.label.getFont().deriveFont(size));
		}
	}
	
	//////////////////// GET/SET METHODS ////////////////////

	/** Get parent LifeLine+ 
	@return parent line */
	public PLifeLinePlus getParentLine() {
		return parentLine;
	}

	/** Set parent LifeLine+
	@return parent line */
	public void setParentLine(PLifeLinePlus parent) {
		parentLine = parent;	
	}

	/** Get title
	@return title of lifeline */
	public String getTitle() {
		return this.title;	
	}
	
	/** Get the user object
	@return user object */
	public MutableTreeNode getUserObject() {
		return this.userObject;
	}
	
	/** Set the default height 
	@param height default height for line */
	public void setDefaultHeight(double height) {
		this.defaultHeight = height;
		this.setExtent();
	}

	/** Get the default height
	@return default line height */
	public double getDefaultHeight() {
		return this.defaultHeight;
	}
	//////////////////// INNER CLASSES ////////////////////

	/** Inner class that represents a summary line when expanding a LifeLine */
	class PSummaryLine extends PNode {
		
		/** Constructor */
		public PSummaryLine() {
			super();
		}
		
		public void paint(PPaintContext aPaintContext) {
			Graphics2D g2 = aPaintContext.getGraphics();
			
			if(this.getPaint() != null) {
				g2.setPaint(this.getPaint());
				
				// draw main line
				g2.fill(new Rectangle2D.Double(0, 0, 
					this.getWidth(), this.getHeight() / 2));
				
				// right triangle
				GeneralPath rightTriangle = new GeneralPath();
				rightTriangle.append(new Line2D.Double(
					this.getWidth() - this.getHeight(), 0,
					this.getWidth(), this.getHeight()), true);
				rightTriangle.append(new Line2D.Double(
					this.getWidth(), this.getHeight(),
					this.getWidth(), 0), true);
				rightTriangle.closePath();
				g2.fill(rightTriangle);	
				
				// left triangle
				GeneralPath leftTriangle = new GeneralPath();
				leftTriangle.append(new Line2D.Double(
					this.getHeight(), 0,
					0, this.getHeight()), true);
				leftTriangle.append(new Line2D.Double(
					0, this.getHeight(),
					0, 0), true);
				leftTriangle.closePath();
				g2.fill(leftTriangle);
			}	
		}
	}
	
}
