/**
ToolTip Event Handler for Piccolo
@author Stephen Kelvin (jazz@gremlin.info)
*/
package at.ac.tuwien.cs.ztimeview.ui;

import edu.umd.cs.piccolo.event.PBasicInputEventHandler;
import edu.umd.cs.piccolo.event.PInputEvent;
import edu.umd.cs.piccolo.event.PInputEventFilter;
import edu.umd.cs.piccolo.PNode;
import edu.umd.cs.piccolo.PCanvas;

import javax.swing.*;
import java.awt.event.*;
import java.awt.*;
import java.awt.geom.Point2D;

public class ToolTipHandler extends PBasicInputEventHandler {
    /**
     * The easiest way to have a tool tip on a PNode is to store the tool tip's text as a client
     * property with this key.
     * The default implementation of {@link #getToolTipText(edu.umd.cs.piccolo.PNode)} uses this
     * key to retrieve the tool tip text.
     * You may override that method if you want to compute the text on the fly.
     * If different locations on the same node should have different tool tip texts you can override
     * {@link #getToolTipText(edu.umd.cs.piccolo.event.PInputEvent) } instead.
     */
    public static final String PICCOLOX_TOOLTIP_PROPERTY_KEY = "PICCOLOX_TOOLTIP";

    /**
     * If there's a mouse entered on a node and the mouse is not moved for initial delay milliseconds,
     * the tool tip is shown.
     */
    private Timer initialTimer;

    /**
     * If there's a mouse exit on a node for which a tooltip is shown and then a mouse
     * entered event on (possibly another) node that has a tool tip then the tool tip
     * is shown immediately if less than the reshow delay milliseconds is between the exited
     * and entered events. The reshowTimer clears the showImmediately flag.
     */
    private Timer reshowTimer;
    
    /**
     * If the tool tip is shown on a node and the mouse has not been moved for at least dismiss delay milliseconds
     * then the tool tip is hidden.
     */
    private Timer dismissTimer;
    
    /**
     * The tool tip's text. Stored to check if the tool tip must be updated when the mouse is moved.
     * (If getToolTipText(PInputEvent) returns different texts for different positions on the node.)
     */
    private String toolTipText;

    /**
     * The tool tip's preferred location as returned from getToolTipLocation(PInputEvent). This is null, unless
     * that method's is overriden. Store to check if the tool tip must be updated when the mouse is moved.
     */
    private Point2D preferredLocation;
    
    /**
     * The piccolo node that the mouse is currently over.
     */
    private PNode insideNode;
    
    /**
     * The last PInput event recorded.
     */
    private PInputEvent inputEvent;
    
    /**
     * If a mouse exits occurs while the tool tip is shown for a node, showImmediately will be true until
     * the reshow timer expires.
     */
    private boolean showImmediately;
    
    /**
     * The popup that shows the tool tip or null if no tool tip is currently shown.
     */
    private Popup tipPopup;

    /**
     * The popup may be display in the same window as the node (if lighweight popups are enabled in the popup factory
     * and the tool tip fits within the window's bounds). tipWindow is null in this case.
     * If however the popup uses a window on its own, it is stored here, so that coordinates can be adjusted.
     */
    private Window tipWindow;
    
    /**
     * The JToolTip used to display the current tool tip. Non-null only if the tip is currently shown.
     * Stored here because on mouse exited on a node it must be checked if the mouse is now over the tool tip itself.
     */
    private JToolTip tip;

    /**
     * True if the tool tip popup is currently shown.
     */
    private boolean tipShowing;
    
    private boolean customInitialDelay;
    private boolean customReshowDelay;
    private boolean customDismissDelay;

    public ToolTipHandler() {
        // setup the event filter
        final PInputEventFilter eventFilter = getEventFilter();
        eventFilter.setAcceptsAlreadyHandledEvents(true);
        eventFilter.setMarksAcceptedEventsAsHandled(false);

        // set up timers for the various delays
        initialTimer = new Timer(750, new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                initialTimerFired();
            }
        });
        initialTimer.setRepeats(false);
        reshowTimer = new Timer(500, new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                reshowTimerFired();
            }
        });
        
        reshowTimer.setRepeats(false);
        dismissTimer = new Timer(4000, new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                dismissTimerFired();
            }
        });
        dismissTimer.setRepeats(false);
    }

    /**
     * In Swing a JComponent registers itself at the ToolTipManager whenever a tool tip text is set on it.
     * We do not have such a close integration with PNodes, so instead of having a separate mouse listener
     * for each component that has a tool tip, there is only a single mouse listener on a PRoot.
     * But now we have to reject all events for nodes that do not have tool tips.
     * @param event
     * @param type
     * @return
     */
    public boolean acceptsEvent(PInputEvent event, int type) {
        boolean accept = super.acceptsEvent(event, type);
        if(accept) {
            final String toolTipText = getToolTipText(event);
            accept = (toolTipText != null);
        }
        return accept;
    }

    /**
     * Called whenever the key stroke for toggling the tool tip is recognized.
     * @param event the key event that caused toggleToolTip() to get called
     */
    private void toggleToolTip(PInputEvent event) {
        PNode toolTipNode = event.getPickedNode();
        if(tipShowing) {
            hideTipWindow();
            insideNode = null;
        }
        else {
            hideTipWindow(); // be safe
            initialTimer.stop();
            reshowTimer.stop();
            dismissTimer.stop();
            insideNode = toolTipNode;
            inputEvent = event;
            if(insideNode != null) {
                toolTipText = getToolTipText(insideNode);
                preferredLocation = new Point2D.Double(10, insideNode.getHeight() + 10); // manual set
                showTipWindow();
            }
        }
    }

    /**
     * Check if the node is in a hierarchy of only visible nodes
     * @param node
     * @return
     */
    private boolean isShowing(PNode node) {
        // @todo should check if there's a camera
        // @todo should check if there's a PCanvas that shows this camera and the PCanvas.isShowing()
        boolean result;
        do {
            result = node.getVisible();
            node = node.getParent();
        } while(result && node != null);

        return result;
    }

    private void showTipWindow() {
        PCanvas canvas = getCanvas();
        if(insideNode == null || !isShowing(insideNode) || !canvas.isShowing())
            return;
        if(ToolTipManager.sharedInstance().isEnabled()) {
            Dimension tipSize;

            Point location = new Point();
            Rectangle screenBounds = canvas.getGraphicsConfiguration().getBounds();
            boolean leftToRight = canvas.getComponentOrientation().isLeftToRight();

            // just to be sure...
            hideTipWindow();

            tip = createToolTip(insideNode);
            tip.setTipText(toolTipText);
            tipSize = tip.getPreferredSize();

            final Point canvasLocationOnScreen = canvas.getLocationOnScreen();
            if(preferredLocation != null) {
                // getPreferredLocation() was overriden in a subclass and returned a non-null
                // position relative to the tooltip node.
                // Convert it to screen coordinates.
                final Point2D preferredLocationGlobal = (Point2D) preferredLocation.clone();
                insideNode.localToGlobal(preferredLocationGlobal);
                
                location.x = canvasLocationOnScreen.x + (int) preferredLocationGlobal.getX();
                location.y = canvasLocationOnScreen.y + (int) preferredLocationGlobal.getY();
            }
            else {
                // Use the default position converted to screen coordinates.
                final Point2D canvasPosition = inputEvent.getCanvasPosition();
                location.x = canvasLocationOnScreen.x + (int) canvasPosition.getX();
                location.y = canvasLocationOnScreen.y + (int) canvasPosition.getY() + 20;
            }
            
            // have not tested this...
            if(!leftToRight) {
                location.x = Math.max(screenBounds.x, location.x - tipSize.width);
            }

            // Fit as much of the tooltip on screen as possible
            if(location.x < screenBounds.x) {
                location.x = screenBounds.x;
            }
            else if(location.x - screenBounds.x + tipSize.width > screenBounds.width) {
                location.x = screenBounds.x + Math.max(0, screenBounds.width - tipSize.width);
            }
            if(location.y < screenBounds.y) {
                location.y = screenBounds.y;
            }
            else if(location.y - screenBounds.y + tipSize.height > screenBounds.height) {
                location.y = screenBounds.y + Math.max(0, screenBounds.height - tipSize.height);
            }
            
            // I am using the default popup factory that is hopefully already configured by Swing's ToolTipManager
            // to use either light, medium or heavyweight popups as its best for the platform we are running on.
            PopupFactory popupFactory = PopupFactory.getSharedInstance();

            tipPopup = popupFactory.getPopup(canvas, tip, location.x, location.y);

            tipPopup.show();

            Window componentWindow = SwingUtilities.windowForComponent(canvas);

            tipWindow = SwingUtilities.windowForComponent(tip);
            if(tipWindow != null && tipWindow != componentWindow) {
                tipWindow.addMouseListener(windowMouseListener);
            }
            else {
                tipWindow = null;
            }

            startDismissTimer();
            tipShowing = true;
        }
    }

    private PCanvas getCanvas() {
        return (PCanvas) inputEvent.getComponent();
    }

    private void handleMouseEntered(PInputEvent event) {
        PNode pickedNode = event.getPickedNode();

        reshowTimer.stop();

        Point2D location = event.getPositionRelativeTo(pickedNode);
        // ensure tooltip shows only in proper place
        if( location.getX() < 0 || location.getX() >= pickedNode.getWidth()
           || location.getY() < 0 || location.getY() >= pickedNode.getHeight()) {
            return;
        }

        if(insideNode != null) {
            initialTimer.stop();
        }

        boolean sameComponent = (insideNode == pickedNode);

        insideNode = pickedNode;
        if(tipPopup != null) {
            inputEvent = event;
            if(showImmediately) {
                String newToolTipText = getToolTipText(event);
                Point2D newPreferredLocation = getToolTipLocation(event);
                boolean sameLoc = (preferredLocation != null) ? preferredLocation.equals(newPreferredLocation)
                                                              : (newPreferredLocation == null);

                if(!sameComponent || !toolTipText.equals(newToolTipText) || !sameLoc) {
                    toolTipText = newToolTipText;
                    preferredLocation = newPreferredLocation;
                    showTipWindow();
                }
            }
            else {
                startInitialTimer();
            }
        }
    }
    
    /**
     * Convenience method for storing the tool tip text in the node's client property
     * {@link #PICCOLOX_TOOLTIP_PROPERTY_KEY PICCOLOX_TOOLTIP_PROPERTY_KEY}.
     * Must not be used if the one of the <code>getToolTipText()</code> methods have been overriden
     * to calculate the tool tip differently.
     * @param node the node that should have the tool tip
     * @param text the tool tip text that should appear for the node
     */
    public static void setToolTipText(PNode node, String text) {
        node.addClientProperty(PICCOLOX_TOOLTIP_PROPERTY_KEY, text);
    }

    /**
     * Used by this ToolTipHandler to get the tool tip text for a node
     * (if {@link #getToolTipText(edu.umd.cs.piccolo.event.PInputEvent) getToolTipText(PInputEvent)}
     * has not been overriden).
     * This default implementation returns the value of the node's client property
     * {@link #PICCOLOX_TOOLTIP_PROPERTY_KEY PICCOLOX_TOOLTIP_PROPERTY_KEY}.
     * A subclass can override this to calculate the tool tip on the fly rather than using the string stored
     * in this client property.
     * @see #setToolTipText(edu.umd.cs.piccolo.PNode, java.lang.String) setToolTipText(PNode, String)
     */
    protected String getToolTipText(PNode node) {
        final Object toolTipProperty = node.getClientProperty(PICCOLOX_TOOLTIP_PROPERTY_KEY);
        String toolTip = (toolTipProperty != null ? toolTipProperty.toString() : null);
        return toolTip;
    }

    /**
     * Used by this ToolTipHandler to get the tool tip text for the node picked by the event.
     * The default implementation calls {@link #getToolTipText(edu.umd.cs.piccolo.PNode) getToolTipText(PNode)}.
     * If a node should have different tool tips depending on the mouse position a subclass can override this method
     * and return a different text depending on the position returned by <code>event.getPosition()</code> (or any
     * other of the event's methods that return positions).
     * @param event the mouse or key event that caused the tool tip to be shown
     * @return the tool tip text or null if no tool tip should be shown
     * (for no tool tip to show getToolTipLocation must also return null, or an empty tool tip will show)
     */
    protected String getToolTipText(PInputEvent event) {
        final String toolTipText = getToolTipText(event.getPickedNode());
        return toolTipText;
    }

    /**
     * Get
     * @param event
     * @return
     */
    protected Point2D getToolTipLocation(PInputEvent event) {
        return null;
    }

    protected JToolTip createToolTip(PNode node) {
        JToolTip tip = new JToolTip();
        return tip;
    }

    /**
     * The tip is currently shown and the mouse has been moved.
     * Now check if the tool tip must be updated.
     */
    private void checkForTipChange(PInputEvent event) {
        String newText = getToolTipText(event);
        Point2D newPreferredLocation = getToolTipLocation(event);

        if(newText == null && newPreferredLocation == null) {
            // no text for this mouse position -> remove tool tip
            toolTipText = null;
            preferredLocation = null;
            inputEvent = null;
            insideNode = null;
            hideTipWindow();
            initialTimer.stop();
            restartReshowTimer();
        }
        else {
            inputEvent = event;
            final boolean textIsEqual = (newText != null && newText.equals(toolTipText)) || newText == null;
            final boolean locationIsEqual = (newPreferredLocation != null && newPreferredLocation.equals(preferredLocation))
                                            || newPreferredLocation == null;
            if(textIsEqual && locationIsEqual) {
                // no changes -> reset dismiss timer

                // tipPopup should be non-null, but just for sure
                if(tipPopup != null) {
                    restartDismissTimer();
                }
                else {
                    restartInitialTimer();
                }
            }
            else {
                // text or location has changed, update tool tip
                toolTipText = newText;
                preferredLocation = newPreferredLocation;
                if(showImmediately) {
                    hideTipWindow();
                    showTipWindow();
                    reshowTimer.stop();
                }
                else {
                    restartInitialTimer();
                }
            }
        }
    }
    
    // If no delays are set explicitly on this ToolTipHandler I am using the delays set at Swing's ToolTipManager,
    // however there is no way to update my own timers immediately when the delays are changed at the ToolTipManager.
    // As a workaround whenever a timer is (re)started the delay is set.
    // Note that calling start() on a timer is ignored when it is already running.
    // Delays can also be set here, so that the defaults from ToolTipManager can be changed.
    
    private void startDismissTimer() {
        dismissTimer.setInitialDelay(getDismissDelay());
        dismissTimer.start();
    }

    private void startInitialTimer() {
        initialTimer.setInitialDelay(getInitialDelay());
        initialTimer.start();
    }

    private void restartInitialTimer() {
        initialTimer.setInitialDelay(getInitialDelay());
        initialTimer.restart();
    }

    private void restartReshowTimer() {
        final int reshowDelay = getReshowDelay();
        if(reshowDelay == 0) {
            // This special case needs to be handled, because else the timer can stopped by the next mouse entry
            // before it fires, even though delay is 0.
            reshowTimer.stop();
            reshowTimerFired();
        }
        else {
            reshowTimer.setInitialDelay(reshowDelay);
            reshowTimer.restart();
        }
    }

    private void restartDismissTimer() {
        dismissTimer.setInitialDelay(getDismissDelay());
        dismissTimer.restart();
    }

    /**
     * If the mouse enters a node and is not moved anymore, it takes this long (in milliseconds) until the tool tip shows.
     * By default this is the same as <code>ToolTipManager.sharedInstance().getInitialDelay()</code>.
     * @return delay in milliseconds from the time the mouse entered a node and is not moved anymore until the tool tip is shown
     */
    public int getInitialDelay() {
        final int initialDelay = (customInitialDelay ? initialTimer.getInitialDelay() : ToolTipManager.sharedInstance().getInitialDelay());
        return initialDelay;
    }

    /**
     * If a tool tip is shown and the mouse exits the node, if it re-enters a node within this long (in milliseconds)
     * the tool tip will reappear immediately (instead of after the initial delay).
     * By default this is the same as <code>ToolTipManager.sharedInstance().getReshowDelay()</code>.
     * @return delay in milliseconds from the time the mouse entered a node and is not moved anymore until the tool tip is shown
     */
    public int getReshowDelay() {
        final int reshowDelay = (customReshowDelay ? reshowTimer.getInitialDelay() : ToolTipManager.sharedInstance().getReshowDelay());
        return reshowDelay;
    }

    /**
     * If a tool tip is shown and the mouse is not moved for this long (in milliseconds) the tool tip is hidden automatically.
     * By default this is the same as <code>ToolTipManager.sharedInstance().getDismissDelay()</code>.
     * @return delay in milliseconds from the time the mouse entered a node and is not moved anymore until the tool tip is shown
     */
    public int getDismissDelay() {
        final int dismissDelay = (customDismissDelay ? dismissTimer.getInitialDelay() : ToolTipManager.sharedInstance().getDismissDelay());
        return dismissDelay;
    }
    
    /**
     * If the mouse enters a node and is not moved anymore, it takes this long (in milliseconds) until the tool tip shows.
     * The value set with this method overrides the value of <code>ToolTipManager.sharedInstance().getInitialDelay()</code>
     * for only this ToolTipHandler instance. It can be set back to that default value by calling <code>resetInitialDelay()</code>.
     * @param milliseconds delay in milliseconds from the time the mouse entered a node and is not moved anymore until the tool tip is shown
     */
    public void setInitialDelay(int milliseconds) {
        customInitialDelay = true;
        initialTimer.setInitialDelay(milliseconds);
    }

    public void resetInitialDelay() {
        customInitialDelay = false;
    }
    
    public void setReshowDelay(int milliseconds) {
        customReshowDelay = true;
        reshowTimer.setInitialDelay(milliseconds);
    }

    public void resetReshowDelay() {
        customReshowDelay = false;
    }
    
    public void setDismissDelay(int milliseconds) {
        customDismissDelay = true;
        dismissTimer.setInitialDelay(milliseconds);
    }

    public void resetDismissDelay() {
        customDismissDelay = false;
    }
    
    private void initialTimerFired() {
        if(insideNode != null && isShowing(insideNode)) {
            // Lazy lookup
            if(toolTipText == null && inputEvent != null) {
                toolTipText = getToolTipText(inputEvent);
                preferredLocation = getToolTipLocation(inputEvent);
            }
            if(toolTipText != null) {
                showImmediately = true;
                showTipWindow();
            }
            else {
                insideNode = null;
                toolTipText = null;
                preferredLocation = null;
                inputEvent = null;
                hideTipWindow();
            }
        }
    }

    private void reshowTimerFired() {
        showImmediately = false;
    }

    private void dismissTimerFired() {
        hideTipWindow();
        initialTimer.stop();
        showImmediately = false;
        insideNode = null;
        inputEvent = null;
    }

    /**
     * A mouse listener installed on the popup window itself (if different from the node's window).
     * Hides the tool tip on mouse pressed and checks if it needs to be hidden on mouse exit.
     */
    private MouseListener windowMouseListener = new MouseAdapter() {
        public void mousePressed(MouseEvent e) {
            handleMousePressed();
        }

        public void mouseExited(MouseEvent event) {
            if(tipWindow != null && event.getSource() == tipWindow) {
                // if we get an exit and have a heavy tipWindow
                // we need to check if it is overlapping the inside component
                final PCanvas canvas = getCanvas();
                Container insideComponentWindow = canvas.getTopLevelAncestor();
                Point location = event.getPoint();
                SwingUtilities.convertPointToScreen(location, tipWindow);

                location.x -= insideComponentWindow.getX();
                location.y -= insideComponentWindow.getY();

                location = SwingUtilities.convertPoint(insideComponentWindow, location, canvas);
                boolean locationIsInside = location.x >= 0 && location.x < insideNode.getWidth()
                                           && location.y >= 0 && location.y < insideNode.getHeight();
                if(!locationIsInside) {
                    hideNow();
                }
            }
        }
    };

    public void mouseEntered(PInputEvent event) {
        handleMouseEntered(event);
    }

    public void mouseExited(PInputEvent event) {
        if(event.getPickedNode() == insideNode) {
            initialTimer.stop();
            if(tipPopup != null) {
                PCanvas canvas = getCanvas();
                Window win = SwingUtilities.getWindowAncestor(canvas);
                if(win != null) {
                    Point2D canvasPosition2D = event.getCanvasPosition();
                    Point canvasPosition = new Point((int) canvasPosition2D.getX(), (int) canvasPosition2D.getY());
                    Point location = SwingUtilities.convertPoint(canvas, canvasPosition, win);
                    Rectangle bounds = canvas.getTopLevelAncestor().getBounds();
                    location.x += bounds.x;
                    location.y += bounds.y;

                    Point loc = new Point(0, 0);
                    SwingUtilities.convertPointToScreen(loc, tip);
                    bounds.x = loc.x;
                    bounds.y = loc.y;
                    bounds.width = tip.getWidth();
                    bounds.height = tip.getHeight();

                    final boolean locationInToolTip = location.x >= bounds.x && location.x < (bounds.x + bounds.width) && location.y >= bounds.y
                                   && location.y < (bounds.y + bounds.height);

                    if(!locationInToolTip) {
                        hideNow();
                    }
                }
            }
        }
    }

    /**
     * Hides the tool tip (if it is currently shown).
     * @param event
     */
    public void mousePressed(PInputEvent event) {
        handleMousePressed();
    }

    public void mouseDragged(PInputEvent event) {
        // nothing to do
    }

    public void mouseMoved(PInputEvent event) {
        if(tipShowing) {
            checkForTipChange(event);
        }
        else if(showImmediately) {
            PNode component = event.getPickedNode();
            toolTipText = getToolTipText(event);
            if(toolTipText != null) {
                preferredLocation = getToolTipLocation(event);
                inputEvent = event;
                insideNode = component;
                reshowTimer.stop();
                showTipWindow();
            }
        }
        else {
            // Lazily lookup the values from within insideTimerAction
            insideNode = event.getPickedNode();
            inputEvent = event;
            toolTipText = null;
            restartInitialTimer();
        }
    }

    public void keyPressed(PInputEvent event) {
        final int keyCode = event.getKeyCode();
        final int modifiers = event.getModifiers();

        final boolean tipShownBefore = tipShowing;

        if(keyCode == KeyEvent.VK_F1 && modifiers == Event.CTRL_MASK) {
            toggleToolTip(event);
        }
        else if(keyCode == KeyEvent.VK_ESCAPE && modifiers == 0) {
            hideToolTip();
        }

        if(tipShowing != tipShownBefore) {
            event.setHandled(true);
        }
    }

    private void handleMousePressed() {
        initialTimer.stop();
        insideNode = null;
        inputEvent = null;
        hideTipWindow();
        showImmediately = false;
    }

    // Should clean up all these variations of hiding the tool tip, however currently the differences are subtle
    // but important, e.g. hideTipWindow() must not set insideNode to null, because it can already hold the
    // next tool tip node.
    
    private void hideTipWindow() {
        if(tipPopup != null) {
            if(tipWindow != null) {
                tipWindow.removeMouseListener(windowMouseListener);
                tipWindow = null;
            }
            tipPopup.hide();
            tipPopup = null;
            tipShowing = false;
            tip.getUI().uninstallUI(tip);
            tip = null;
            dismissTimer.stop();
        }
    }

    private void hideToolTip() {
        hideTipWindow();
        preferredLocation = null;
        insideNode = null;
    }

    private void hideNow() {
        initialTimer.stop();
        insideNode = null;
        toolTipText = null;
        inputEvent = null;
        hideTipWindow();
        restartReshowTimer();
    }
} 